# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import numpy as np

from model_compression_toolkit.core.common.logger import Logger
from model_compression_toolkit.core.common.constants import FOUND_TF

from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.quantizers_infrastructure import QuantizationTarget
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import mark_quantizer


if FOUND_TF:
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.keras.quantizers.activation_inferable_quantizers.activation_symmetric_inferable_quantizer \
        import ActivationSymmetricInferableQuantizer

    @mark_quantizer(quantization_target=QuantizationTarget.Activation,
                    quantization_method=[QuantizationMethod.POWER_OF_TWO],
                    quantizer_type=None)
    class ActivationPOTInferableQuantizer(ActivationSymmetricInferableQuantizer):
        """
        Class for quantizing activations using power-of-two quantizer
        """

        def __init__(self,
                     num_bits: int,
                     threshold: List[float],
                     signed: bool):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                threshold: threshold for quantizing activations
                signed: whether or not to use signed quantization
            """
            # Call the superclass constructor with the given parameters, along with the target of Activation
            # quantization
            super(ActivationPOTInferableQuantizer, self).__init__(num_bits=num_bits,
                                                                  threshold=threshold,
                                                                  signed=signed)

            is_threshold_pot = np.all([int(np.log2(x)) == np.log2(x) for x in self.threshold.flatten()])
            assert is_threshold_pot, f'Expected threshold to be power of 2 but is {self.threshold}'

else:
    class ActivationPOTInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            Logger.error('Installing tensorflow and tensorflow_model_optimization is mandatory '
                         'when using ActivationPOTInferableQuantizer. '
                         'Could not find Tensorflow package.')



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from . import base_matcher


class BaseNodeMatcher(base_matcher.BaseMatcher):
    """
    Base class for matchers that match a node in the graph.
    An extension of logics AND, OR and NOT are implemented.
    """

    def __and__(self, other):
        """
        Return a matcher to check the logic AND of two BaseNodeMatcher on an object.
        """
        return NodeAndMatcher(self, other)

    def __or__(self, other):
        """
        Return a matcher to check the logic OR of two BaseNodeMatcher on an object.
        """
        return NodeOrMatcher(self, other)

    def logic_not(self):
        """
        Return a matcher to check the logic NOT of the BaseNodeMatcher on an object.
        """
        return NodeNotMatcher(self)


class NodeAndMatcher(BaseNodeMatcher):
    """
    Matcher to check the logic AND of two node matchers on an object.
    """

    def __init__(self, matcher_a: BaseNodeMatcher, matcher_b: BaseNodeMatcher):
        self.matcher_a = matcher_a
        self.matcher_b = matcher_b

    def apply(self, input_object) -> bool:
        return self.matcher_a.apply(input_object) and self.matcher_b.apply(input_object)


class NodeOrMatcher(BaseNodeMatcher):
    """
    Matcher to check the logic OR of two node matchers on an object.
    """

    def __init__(self, matcher_a: BaseNodeMatcher, matcher_b: BaseNodeMatcher):
        self.matcher_a = matcher_a
        self.matcher_b = matcher_b

    def apply(self, input_object) -> bool:
        return self.matcher_a.apply(input_object) or self.matcher_b.apply(input_object)


class NodeAnyMatcher(BaseNodeMatcher):
    """
    Matcher to return always True, to any given input object.
    """

    def apply(self, input_object) -> bool:
        return True


class NodeNotMatcher(BaseNodeMatcher):
    """
    Matcher to check the logic AND of two node matchers on an object.
    """

    def __init__(self, matcher_a):
        self.matcher_a = matcher_a

    def apply(self, input_object) -> bool:
        return not self.matcher_a.apply(input_object)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from abc import abstractmethod
from enum import Enum
from typing import Union, List, Any
from inspect import signature

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import Logger

from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import BaseInferableQuantizer, \
    QuantizationTarget
from model_compression_toolkit.quantizers_infrastructure.trainable_infrastructure.common.trainable_quantizer_config import \
    TrainableQuantizerActivationConfig, TrainableQuantizerWeightsConfig
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.constants import QUANTIZATION_METHOD, \
    QUANTIZATION_TARGET


VAR = 'var'
GROUP = 'group'

class VariableGroup(Enum):
    """
    An enum for choosing trainable variable group
    0. WEIGHTS
    1. QPARAMS
    """
    WEIGHTS = 0
    QPARAMS = 1


class BaseTrainableQuantizer(BaseInferableQuantizer):
    def __init__(self,
                 quantization_config: Union[TrainableQuantizerActivationConfig, TrainableQuantizerWeightsConfig]):
        """
        This class is a base quantizer which validates the provided quantization config and defines an abstract function which any quantizer needs to implment.

        Args:
            quantization_config: quantizer config class contains all the information about the quantizer configuration.
        """

        # verify the quantizer class that inherits this class only has a config argument and key-word arguments
        for i, (k, v) in enumerate(self.get_sig().parameters.items()):
            if i == 0:
                if v.annotation not in [TrainableQuantizerWeightsConfig, TrainableQuantizerActivationConfig]:
                    common.Logger.error(f"First parameter must be either TrainableQuantizerWeightsConfig or TrainableQuantizerActivationConfig")  # pragma: no cover
            elif v.default is v.empty:
                common.Logger.error(f"Parameter {k} doesn't have a default value")  # pragma: no cover

        super(BaseTrainableQuantizer, self).__init__()
        self.quantization_config = quantization_config

        # Inherited class should be decorated with @mark_quantizer decorator, and define the following static properties
        static_quantization_method = getattr(self, QUANTIZATION_METHOD, None)
        static_quantization_target = getattr(self, QUANTIZATION_TARGET, None)

        if static_quantization_method is None or static_quantization_target is None:
            Logger.error("A quantizer class that inherit from BaseTrainableQuantizer is not defined appropriately."
                         "Either it misses the @mark_quantizer decorator or the decorator is not used correctly.")

        if static_quantization_target == QuantizationTarget.Weights:
            self.validate_weights()
            if self.quantization_config.weights_quantization_method not in static_quantization_method:
                common.Logger.error(
                    f'Quantization method mismatch expected: {static_quantization_method} and got  {self.quantization_config.weights_quantization_method}')
        elif static_quantization_target == QuantizationTarget.Activation:
            self.validate_activation()
            if self.quantization_config.activation_quantization_method not in static_quantization_method:
                common.Logger.error(
                    f'Quantization method mismatch expected: {static_quantization_method} and got  {self.quantization_config.activation_quantization_method}')
        else:
            common.Logger.error(
                f'Unknown Quantization Part:{static_quantization_target}')  # pragma: no cover

        self.quantizer_parameters = {}

    @classmethod
    def get_sig(cls):
        return signature(cls)

    def initialize_quantization(self,
                                tensor_shape,
                                name: str,
                                layer):
        """
        This initializes the quantizer parameters given the parameter name and shape.

        Args:
            tensor_shape:  tensor shape
            name: tensor name
            layer: layer to quantized

        Returns: None

        """
        raise NotImplemented  # pragma: no cover

    def __call__(self,
                 input2quantize,
                 training: bool):
        """
        Quantize a tensor.

        Args:
            input2quantize: Input tensor to quantize.
            training: Whether the graph is in training mode.

        Returns:
            The quantized tensor.
        """
        raise NotImplemented  # pragma: no cover

    def activation_quantization(self) -> bool:
        """

        Returns: A boolean stating is this activation quantizer

        """
        return isinstance(self.quantization_config, TrainableQuantizerActivationConfig)

    def weights_quantization(self) -> bool:
        """

        Returns: A boolean stating is this weights quantizer

        """
        return isinstance(self.quantization_config, TrainableQuantizerWeightsConfig)

    def validate_weights(self) -> None:
        """
        This function validates the quantization config compared with its parameters.


        """
        if self.activation_quantization() or not self.weights_quantization():
            common.Logger.error(f'Expect weight quantization got activation')

    def validate_activation(self) -> None:
        """
        This function validates the quantization config compared with its parameters.

        """
        if not self.activation_quantization() or self.weights_quantization():
            common.Logger.error(f'Expect activation quantization got weight')

    def convert2inferable(self) -> BaseInferableQuantizer:
        """
        Convert quantizer to inferable quantizer.

        Returns:
            BaseInferableQuantizer object.
        """
        raise NotImplemented  # pragma: no cover

    def add_quantizer_variable(self, name: str, variable: Any, group: VariableGroup = VariableGroup.WEIGHTS):
        """
        Add a quantizer variable to quantizer_parameters dictionary
        """
        self.quantizer_parameters.update({name: {VAR: variable, GROUP: group}})

    def get_quantizer_variable(self, name: str) -> Any:
        """
        Get a quantizer variable by name

        Args:
            name: variable name

        Returns:
            trainable variable
        """
        if name in self.quantizer_parameters:
            return self.quantizer_parameters[name][VAR]
        else:
            common.Logger.error(f'Variable {name} is not exist in quantizers parameters!') # pragma: no cover


    @abstractmethod
    def get_trainable_variables(self, group: VariableGroup) -> List[Any]:
        """
        Get trainable parameters with specific group from quantizer

        Args:
            group: Enum of variable group

        Returns:
            List of trainable variables
        """
        raise NotImplemented  # pragma: no cover


# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# Inferable keras quantizer signature parameters:
NUM_BITS = 'num_bits'
SIGNED = 'signed'
THRESHOLD = 'threshold'
PER_CHANNEL = 'per_channel'
MIN_RANGE = 'min_range'
MAX_RANGE = 'max_range'
CHANNEL_AXIS = 'channel_axis'
INPUT_RANK = 'input_rank'
CLUSTER_CENTERS = 'cluster_centers'


# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np

from model_compression_toolkit.core.common.constants import FOUND_TORCH
from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer \
    import mark_quantizer, \
    QuantizationTarget
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.constants \
    import MULTIPLIER_N_BITS, EPS

if FOUND_TORCH:
    import torch
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.pytorch.quantizer_utils \
        import to_torch_tensor, get_working_device, lut_quantizer
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.pytorch.quantizers \
        .base_lut_symmetric_inferable_quantizer import BaseLUTSymmetricInferableQuantizer


    @mark_quantizer(quantization_target=QuantizationTarget.Weights,
                    quantization_method=[QuantizationMethod.LUT_SYM_QUANTIZER],
                    quantizer_type=None)
    class WeightsLUTSymmetricInferableQuantizer(BaseLUTSymmetricInferableQuantizer):
        """
        Class for quantizing weights using a lut symmetric quantizer
        """

        def __init__(self,
                     num_bits: int,
                     cluster_centers: np.ndarray,
                     threshold: np.ndarray,
                     per_channel: bool,
                     channel_axis: int = None,
                     multiplier_n_bits: int = MULTIPLIER_N_BITS,
                     eps: float = EPS):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                cluster_centers: the cluster centers to assign the weights
                threshold: threshold for quantizing weights
                per_channel: whether to use per-channel quantization
                channel_axis: Axis of input to apply per-channel quantization on
                multiplier_n_bits: Number of bits that determines the quantization range
                eps: Small value for numerical stability in division
            """

            super(WeightsLUTSymmetricInferableQuantizer, self).__init__(threshold=threshold,
                                                                        num_bits=num_bits,
                                                                        cluster_centers=cluster_centers,
                                                                        signed=True,
                                                                        multiplier_n_bits=multiplier_n_bits,
                                                                        eps=eps)

            if per_channel:
                assert channel_axis is not None, f'Channel axis is missing in per channel quantization'
                assert len(
                    threshold) >= 1, f'In per-channel quantization threshold should be of length >= 1 but is ' \
                                     f'{len(threshold)}'
            else:
                assert len(
                    threshold) == 1, f'In per-tensor quantization threshold should be of length 1 but is ' \
                                     f'{len(threshold)}'

            self.per_channel = per_channel
            self.channel_axis = channel_axis

            self.threshold = to_torch_tensor(self.threshold).to(get_working_device())
            self.cluster_centers = to_torch_tensor(self.cluster_centers).to(get_working_device())

        def __call__(self, inputs: torch.Tensor) -> torch.Tensor:
            """
            Quantize the given inputs using the quantizer parameters.

            Args:
                inputs: input tensor to quantize

            Returns:
                quantized tensor.
            """
            inputs.requires_grad = False
            return lut_quantizer(inputs, cluster_centers=self.cluster_centers, signed=True,
                                 threshold=self.threshold, multiplier_n_bits=self.multiplier_n_bits, eps=self.eps)


else:
    class WeightsLUTSymmetricInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing torch is mandatory '
                            'when using WeightsLUTSymmetricInferableQuantizer. '
                            'Could not find torch package.')



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch
import torch.nn as nn
from typing import Dict
import numpy as np

from model_compression_toolkit import quantizers_infrastructure as qi
from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.gptq.common.gptq_config import RoundingType
from model_compression_toolkit.gptq.pytorch.quantizer.base_pytorch_gptq_quantizer import \
    BasePytorchGPTQTrainableQuantizer
from model_compression_toolkit.core.pytorch.utils import to_torch_tensor, torch_tensor_to_numpy
from model_compression_toolkit.gptq.pytorch.quantizer import quant_utils as qutils
from model_compression_toolkit.gptq.common.gptq_constants import SOFT_ROUNDING_GAMMA, SOFT_ROUNDING_ZETA, AUXVAR
from model_compression_toolkit.gptq.pytorch.quantizer.quant_utils import fix_range_to_include_zero
from model_compression_toolkit.quantizers_infrastructure import TrainableQuantizerWeightsConfig
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import \
    mark_quantizer
from model_compression_toolkit.quantizers_infrastructure.trainable_infrastructure.common.base_trainable_quantizer import \
    VariableGroup
from model_compression_toolkit.core.common.constants import RANGE_MAX, RANGE_MIN
from model_compression_toolkit.qat.common.constants import FQ_MIN, FQ_MAX

def soft_rounding_unifrom_quantizer(input_tensor: torch.Tensor,
                                    auxvar_tensor: torch.Tensor,
                                    min_range: torch.Tensor,
                                    max_range: torch.Tensor,
                                    num_bits: int) -> torch.Tensor:
    """
    Quantize a tensor uniformly for GPTQ quantizers.

    Args:
        input_tensor: Tensor to quantize. values of this tensor are not changed during gptq.
        auxvar_tensor: Tensor that manifests the bit shift of the quantized weights due to gptq training.
        min_range: Tensor with min values to compute the delta grid.
        max_range: Tensor with max values to compute the delta grid.
        num_bits: Num of bits to use.

    Returns:
        A quantized tensor.
    """
    # adjusts the quantization range so the quantization grid includes zero.
    min_range, max_range = fix_range_to_include_zero(min_range, max_range, num_bits)
    delta = qutils.calculate_delta_uniform(max_range, min_range, num_bits)
    with torch.no_grad():
        input_tensor_int = torch.floor(input_tensor / delta)
    tensor_q = input_tensor_int + auxvar_tensor
    return delta * qutils.ste_clip(tensor_q,
                                   min_val=0,
                                   max_val=2 ** num_bits - 1)


@mark_quantizer(quantization_target=qi.QuantizationTarget.Weights,
                quantization_method=[QuantizationMethod.UNIFORM],
                quantizer_type=RoundingType.SoftQuantizer)
class UniformSoftRoundingGPTQ(BasePytorchGPTQTrainableQuantizer):
    """
    Trainable uniform quantizer to optimize the rounding of the quantized values using a soft quantization method.
    """

    def __init__(self,
                 quantization_config: TrainableQuantizerWeightsConfig,
                 quantization_parameter_learning: bool = False):
        """
        Construct a Pytorch model that utilize a fake weight quantizer of soft-quantizer for symmetric quantizer.

        Args:
            quantization_config: Trainable weights quantizer config.
            quantization_parameter_learning (Bool): Whether to learn the min/max ranges or not
        """

        super().__init__(quantization_config)
        self.num_bits = quantization_config.weights_n_bits
        self.per_channel = quantization_config.weights_per_channel_threshold

        self.min_values = quantization_config.weights_quantization_params[RANGE_MIN]
        self.max_values = quantization_config.weights_quantization_params[RANGE_MAX]

        self.quantization_axis = quantization_config.weights_channels_axis
        self.quantization_parameter_learning = quantization_parameter_learning

        # gamma and zeta are stretch parameters for computing the rectified sigmoid function.
        # See: https://arxiv.org/pdf/2004.10568.pdf
        self.gamma = SOFT_ROUNDING_GAMMA
        self.zeta = SOFT_ROUNDING_ZETA

    def initialize_quantization(self,
                                tensor_shape: torch.Size,
                                name: str,
                                layer: qi.PytorchQuantizationWrapper):
        """
        Add quantizer parameters to the quantizer parameters dictionary

        Args:
            tensor_shape: tensor shape of the quantized tensor.
            name: Tensor name.
            layer: Layer to quantize.
        """

        # Add min and max variables to layer.
        if self.per_channel:
            min_values = to_torch_tensor(self.min_values)
            max_values = to_torch_tensor(self.max_values)
        else:
            min_values = torch.tensor(self.min_values)
            max_values = torch.tensor(self.max_values)

        layer.register_parameter(name+"_"+FQ_MIN, nn.Parameter(min_values, requires_grad=self.quantization_parameter_learning))
        layer.register_parameter(name+"_"+FQ_MAX, nn.Parameter(max_values, requires_grad=self.quantization_parameter_learning))

        w = layer.layer.weight
        delta = qutils.calculate_delta_uniform(max_values, min_values, self.num_bits)
        w_clipped_normed = torch.clip(w / delta, 0, 2 ** self.num_bits - 1)
        rest = w_clipped_normed - torch.floor(w_clipped_normed)  # rest of rounding [0, 1)
        alpha = -torch.log((self.zeta - self.gamma) / (rest - self.gamma) - 1)  # => sigmoid(alpha) = rest
        layer.register_parameter(f"{name}_{AUXVAR}", nn.Parameter(alpha, requires_grad=True))

        # Save the quantizer parameters
        self.add_quantizer_variable(FQ_MIN, layer.get_parameter(name+"_"+FQ_MIN), VariableGroup.QPARAMS)
        self.add_quantizer_variable(FQ_MAX, layer.get_parameter(name+"_"+FQ_MAX), VariableGroup.QPARAMS)
        self.add_quantizer_variable(AUXVAR, layer.get_parameter(f"{name}_{AUXVAR}"), VariableGroup.WEIGHTS)


    def get_soft_targets(self) -> torch.Tensor:
        """
        Computes the rectified sigmoid function for the quantization target parameters.

        Returns:
            A tensor with the soft rounding targets values.

        """
        scaled_sigmoid = torch.sigmoid(self.get_quantizer_variable(AUXVAR)) * (self.zeta - self.gamma) + self.gamma
        return torch.clip(scaled_sigmoid, min=0, max=1)

    def get_quant_config(self) -> Dict[str, np.ndarray]:
        """
        Returns the config used to edit NodeQuantizationConfig after GPTQ retraining

        Returns:
            A dictionary of attributes the quantize_config retraining has changed during GPTQ retraining.
            Keys must match NodeQuantizationConfig attributes

        """
        min_values = torch_tensor_to_numpy(self.get_quantizer_variable(FQ_MIN))
        max_values = torch_tensor_to_numpy(self.get_quantizer_variable(FQ_MAX))
        return {RANGE_MIN:  min_values,
                RANGE_MAX:  max_values}

    def __call__(self,
                 inputs: nn.Parameter,
                 training: bool) -> torch.Tensor:
        """
        Quantize a tensor.

        Args:
            inputs: Input tensor to quantize.
            training: whether in training mode or not

        Returns:
            quantized tensor
        """
        auxvar = self.get_quantizer_variable(AUXVAR)
        min_range = self.get_quantizer_variable(FQ_MIN)
        max_range = self.get_quantizer_variable(FQ_MAX)

        #####################################################
        # Soft Rounding
        #####################################################
        aux_var = self.get_soft_targets()
        if not training:
            aux_var = (aux_var >= 0.5).to(auxvar.dtype)

        #####################################################
        # Quantized Input
        #####################################################
        q_tensor = soft_rounding_unifrom_quantizer(input_tensor=inputs,
                                                   auxvar_tensor=aux_var,
                                                   min_range=min_range,
                                                   max_range=max_range,
                                                   num_bits=self.num_bits)


        return q_tensor



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Any, List, Tuple
import networkx as nx

from model_compression_toolkit.core.common.logger import Logger


class DirectedBipartiteGraph(nx.DiGraph):
    """
    Directed Bipartite graph representation.
    """

    def __init__(self,
                 name: str,
                 a_nodes: List[Any],
                 b_nodes: List[Any],
                 edges_ab: List[Tuple[Any, Any]],
                 edges_ba: List[Tuple[Any, Any]],
                 **attr):
        """
        Args:
            name: A name to reference the graph with.
            a_nodes: A set of objects to define as the A-side in the bipartite graph.
            b_nodes: A set of objects to define as the B-side in the bipartite graph.
            edges_ab: Edges from side A to side B.
            edges_ba: Edges from side B to side A.
            **attr: Attributes to add to graph as key=value pairs.
        """

        super().__init__(**attr)

        self.name = name

        self.a_nodes = set()
        self.b_nodes = set()

        self.add_nodes_from(a_nodes, bipartite=0)
        self.add_nodes_from(b_nodes, bipartite=1)
        self._update_nodes_sets()

        # Check edges validity
        self._verify_edges(edges_ab)
        self._verify_edges(edges_ba)

        self.add_edges_from(edges_ab)
        self.add_edges_from(edges_ba)

    def _update_nodes_sets(self):
        """
        Updates the class members of side A and side B nodes.
        """
        self.a_nodes = {n for n, d in self.nodes(data=True) if d["bipartite"] == 0}
        self.b_nodes = set(self) - self.a_nodes

    def _verify_edges(self, edges_list: List[Tuple[Any, Any]]):
        """
        Verifies bipartite correctness of a set of edges to add to the graph.
        If there is an edge in the list that violates the bipartite property - an Exception is raised.

        Args:
            edges_list: A list of edges to verify their correction.
        """
        for n1, n2 in edges_list:
            if n1 in self.a_nodes and n2 in self.a_nodes:
                Logger.critical(f"Can't add an edge {(n1, n2)} between two nodes in size A of a bipartite graph.")
            if n1 in self.b_nodes and n2 in self.b_nodes:
                Logger.critical(f"Can't add an edge {(n1, n2)} between two nodes in size B of a bipartite graph.")

    def add_nodes_to_a(self, new_nodes: List[Any]):
        """
        Add a set of nodes to side A of the bipartite graph.

        Args:
            new_nodes: New nodes to add to side A.
        """
        self.add_nodes_from(new_nodes, bipartite=0)
        self._update_nodes_sets()

    def add_nodes_to_b(self, new_nodes: List[Any]):
        """
        Add a set of nodes to side B of the bipartite graph.

        Args:
            new_nodes: New nodes to add to side B.
        """
        self.add_nodes_from(new_nodes, bipartite=1)
        self._update_nodes_sets()

    def add_edges(self, new_edges: List[Tuple[Any, Any]]):
        """
        Add a set of edges to the bipartite graph.

        Args:
            new_edges: New edges to add to side B.
        """
        self._verify_edges(new_edges)
        self.add_edges_from(new_edges)



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np

from model_compression_toolkit.core.common.constants import FOUND_TORCH
from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import mark_quantizer

if FOUND_TORCH:
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.pytorch.quantizers.base_pytorch_inferable_quantizer import \
        BasePyTorchInferableQuantizer


    @mark_quantizer(quantization_target=None,
                    quantization_method=[QuantizationMethod.UNIFORM],
                    quantizer_type=None)
    class BaseUniformInferableQuantizer(BasePyTorchInferableQuantizer):

        def __init__(self,
                     num_bits: int,
                     min_range: np.ndarray,
                     max_range: np.ndarray):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                min_range: min quantization range for quantizing
                max_range: max quantization range for quantizing
            """

            super(BaseUniformInferableQuantizer, self).__init__()
            self.num_bits = num_bits
            self.min_range = min_range
            self.max_range = max_range
            self.min_quantized_domain = 0
            self.max_quantized_domain = 2 ** num_bits - 1


else:
    class BaseUniformInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing torch is mandatory '
                            'when using BaseUniformInferableQuantizer. '
                            'Could not find torch package.')



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Callable

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.constants import PYTORCH, FOUND_TORCH
from model_compression_toolkit.core.common.target_platform import TargetPlatformCapabilities
from model_compression_toolkit.core.common.mixed_precision.kpi_tools.kpi import KPI
from model_compression_toolkit import CoreConfig
from model_compression_toolkit.core.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfigV2
from model_compression_toolkit.core.runner import core_runner, _init_tensorboard_writer
from model_compression_toolkit.ptq.runner import ptq_runner
from model_compression_toolkit.core.exporter import export_model
from model_compression_toolkit.core.analyzer import analyzer_model_quantization


if FOUND_TORCH:
    from model_compression_toolkit.core.pytorch.default_framework_info import DEFAULT_PYTORCH_INFO
    from model_compression_toolkit.core.pytorch.pytorch_implementation import PytorchImplementation
    from model_compression_toolkit.core.pytorch.constants import DEFAULT_TP_MODEL
    from torch.nn import Module
    from model_compression_toolkit.exporter.model_wrapper.pytorch.builder.fully_quantized_model_builder import get_exportable_pytorch_model
    from model_compression_toolkit import get_target_platform_capabilities

    DEFAULT_PYTORCH_TPC = get_target_platform_capabilities(PYTORCH, DEFAULT_TP_MODEL)

    def pytorch_post_training_quantization_experimental(in_module: Module,
                                                        representative_data_gen: Callable,
                                                        target_kpi: KPI = None,
                                                        core_config: CoreConfig = CoreConfig(),
                                                        target_platform_capabilities: TargetPlatformCapabilities = DEFAULT_PYTORCH_TPC,
                                                        new_experimental_exporter: bool = False):
        """
        Quantize a trained Pytorch module using post-training quantization.
        By default, the module is quantized using a symmetric constraint quantization thresholds
        (power of two) as defined in the default TargetPlatformCapabilities.
        The module is first optimized using several transformations (e.g. BatchNormalization folding to
        preceding layers). Then, using a given dataset, statistics (e.g. min/max, histogram, etc.) are
        being collected for each layer's output (and input, depends on the quantization configuration).
        Thresholds are then being calculated using the collected statistics and the module is quantized
        (both coefficients and activations by default).
        If gptq_config is passed, the quantized weights are optimized using gradient based post
        training quantization by comparing points between the float and quantized modules, and minimizing the
        observed loss.

        Args:
            in_module (Module): Pytorch module to quantize.
            representative_data_gen (Callable): Dataset used for calibration.
            target_kpi (KPI): KPI object to limit the search of the mixed-precision configuration as desired.
            core_config (CoreConfig): Configuration object containing parameters of how the model should be quantized, including mixed precision parameters.
            target_platform_capabilities (TargetPlatformCapabilities): TargetPlatformCapabilities to optimize the PyTorch model according to.
            new_experimental_exporter (bool): Whether exporting the quantized model using new exporter or not (in progress. Avoiding it for now is recommended).

        Returns:
            A quantized module and information the user may need to handle the quantized module.

        Examples:

            Import a Pytorch module:

            >>> from torchvision import models
            >>> module = models.mobilenet_v2()

            Create a random dataset generator, for required number of calibration iterations (num_calibration_batches):
            In this example a random dataset of 10 batches each containing 4 images is used.

            >>> import numpy as np
            >>> num_calibration_batches = 10
            >>> def repr_datagen():
            >>>     for _ in range(num_calibration_batches):
            >>>         yield [np.random.random((4, 3, 224, 224))]

            Import MCT and pass the module with the representative dataset generator to get a quantized module
            Set number of clibration iterations to 1:

            >>> import model_compression_toolkit as mct
            >>> quantized_module, quantization_info = mct.pytorch_post_training_quantization_experimental(module, repr_datagen)

        """

        if core_config.mixed_precision_enable:
            if not isinstance(core_config.mixed_precision_config, MixedPrecisionQuantizationConfigV2):
                common.Logger.error("Given quantization config to mixed-precision facade is not of type "
                                    "MixedPrecisionQuantizationConfigV2. Please use "
                                    "pytorch_post_training_quantization API, or pass a valid mixed precision "
                                    "configuration.")  # pragma: no cover

            common.Logger.info("Using experimental mixed-precision quantization. "
                               "If you encounter an issue please file a bug.")

        tb_w = _init_tensorboard_writer(DEFAULT_PYTORCH_INFO)

        fw_impl = PytorchImplementation()

        tg, bit_widths_config = core_runner(in_model=in_module,
                                            representative_data_gen=representative_data_gen,
                                            core_config=core_config,
                                            fw_info=DEFAULT_PYTORCH_INFO,
                                            fw_impl=fw_impl,
                                            tpc=target_platform_capabilities,
                                            target_kpi=target_kpi,
                                            tb_w=tb_w)

        tg = ptq_runner(tg, representative_data_gen, core_config, DEFAULT_PYTORCH_INFO, fw_impl, tb_w)

        if core_config.debug_config.analyze_similarity:
            analyzer_model_quantization(representative_data_gen,
                                        tb_w,
                                        tg,
                                        fw_impl,
                                        DEFAULT_PYTORCH_INFO)

        if new_experimental_exporter:
            Logger.warning('Using new experimental exported models. '
                           'Please do not use unless you are familiar with what you are doing')

            return get_exportable_pytorch_model(tg)

        quantized_model, user_info = export_model(tg,
                                                  DEFAULT_PYTORCH_INFO,
                                                  fw_impl,
                                                  tb_w,
                                                  bit_widths_config)

        return quantized_model, user_info

else:
    # If torch is not installed,
    # we raise an exception when trying to use these functions.
    def pytorch_post_training_quantization_experimental(*args, **kwargs):
        Logger.critical('Installing Pytorch is mandatory '
                        'when using pytorch_post_training_quantization_experimental. '
                        'Could not find the torch package.')  # pragma: no cover



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Set, Tuple, Any
import numpy as np


class ActivationMemoryTensor:
    """
    A representation of an activation output tensor of a model's layer.
    """

    def __init__(self, shape: Tuple[Any], node_name: str, node_output_index: int, init_size_to_zero: bool = False):
        """
        Args:
            shape: The shape of the activation tensor.
            node_name: The name of the node which its output is represented by the object.
            node_output_index: The index of this tensor in the node's outputs list.
            init_size_to_zero: Whether to initialize the memory tensor size to 0 or not.
        """

        self.shape = shape[1:]  # remove batch size (first element) from output shape
        # The total size of a tensor is considered to be the number of elements in the tensor
        self.total_size = self._get_tensor_total_size() if not init_size_to_zero else 0

        self.node_name = node_name
        self.node_output_index = node_output_index

    def _get_tensor_total_size(self) -> np.ndarray:
        """
        Returns: The total number of parameters in an activation tensor.
        """

        assert all([x is not None for x in self.shape])
        return np.prod(self.shape)


class MemoryElements:
    """
    An object which represents a set of activation tensors and their memory size.
    """

    def __init__(self, elements: Set[ActivationMemoryTensor], total_size: float):
        """
        Args:
            elements: A set of  ActivationMemoryTensor (the memory elements)
            total_size: The total number of parameters of the given tensors.
        """
        self.elements = elements
        self.total_size = total_size

    def add_element(self, new_element: ActivationMemoryTensor):
        """
        Adding an element to the set.

        Args:
            new_element: The element to add.

        """
        self.elements.add(new_element)
        self.total_size += new_element.total_size

    def add_elements_set(self, new_elements_set: Set[ActivationMemoryTensor]):
        """
        Adding a set of elements to the set.

        Args:
            new_elements_set: The elements to add.

        """
        self.elements.update(new_elements_set)
        self.total_size += sum([e.total_size for e in new_elements_set])

    def __eq__(self, other) -> bool:
        """
        Overrides the class equality method.
        Two MemoryElements objects are equal if they contain the same elements.

        Args:
            other: An object to compare the current object to.

        Returns: True if the two objects are equal. False otherwise.

        """
        if isinstance(other, MemoryElements):
            # MemoryElements are equal if they contain the exact same elements sets
            return self.elements == other.elements
        return False

    def __hash__(self):
        return hash((frozenset(self.elements)))

    def __copy__(self):
        """
        Overrides the class copy method.
        Creates a new set with the same elements that are in the copied object.

        Returns: A new MemoryElements object with a copied set of elements.

        """
        return MemoryElements({elm for elm in self.elements}, self.total_size)



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from enum import Enum
from typing import Callable, Any, Dict
from model_compression_toolkit.core.common.defaultdict import DefaultDict
from model_compression_toolkit.core import common
from model_compression_toolkit.gptq.common.gptq_constants import QUANT_PARAM_LEARNING_STR, MAX_LSB_STR, REG_DEFAULT


class RoundingType(Enum):
    """
    An enum for choosing the GPTQ rounding methods
    0. STRAIGHT-THROUGH ESTIMATOR
    1. SoftQuantizer
    """
    STE = 0
    SoftQuantizer = 1


class GPTQHessianWeightsConfig:
    """
    Configuration to use for computing the Hessian-based weights for GPTQ loss metric.
    """

    def __init__(self,
                 hessians_num_samples: int = 16,
                 norm_weights: bool = True,
                 log_norm: bool = True,
                 scale_log_norm: bool = False,
                 hessians_n_iter: int = 50):

        """
        Initialize a GPTQHessianWeightsConfig.
        Args:
            hessians_num_samples (int): Number of samples to use for computing the Hessian-based weights.
            norm_weights (bool): Whether to normalize the returned weights (to get values between 0 and 1).
            log_norm (bool): Whether to use log normalization to the GPTQ Hessian-based weights.
            scale_log_norm (bool): Whether to scale the final vector of the Hessian weights.
            hessians_n_iter (int): Number of random iterations to run Hessian approximation for GPTQ weights.
        """

        self.hessians_num_samples = hessians_num_samples
        self.norm_weights = norm_weights
        self.log_norm = log_norm
        self.scale_log_norm = scale_log_norm
        self.hessians_n_iter = hessians_n_iter


class GradientPTQConfig:
    """
    Configuration to use for quantization with GradientPTQ (experimental).
    """

    def __init__(self, n_iter: int,
                 optimizer: Any,
                 optimizer_rest: Any = None,
                 loss: Callable = None,
                 log_function: Callable = None,
                 train_bias: bool = True,
                 rounding_type: RoundingType = RoundingType.SoftQuantizer,
                 use_hessian_based_weights: bool = True,
                 optimizer_quantization_parameter: Any = None,
                 optimizer_bias: Any = None,
                 regularization_factor: float = REG_DEFAULT,
                 hessian_weights_config: GPTQHessianWeightsConfig = GPTQHessianWeightsConfig(),
                 gptq_quantizer_params_override: Dict[str, Any] = None):
        """
        Initialize a GradientPTQConfig.

        Args:
            n_iter (int): Number of iterations to train.
            optimizer (Any): Optimizer to use.
            optimizer_rest (Any): Optimizer to use for bias and quantizer parameters.
            loss (Callable): The loss to use. should accept 6 lists of tensors. 1st list of quantized tensors, the 2nd list is the float tensors,
             the 3rd is a list of quantized weights, the 4th is a list of float weights, the 5th and 6th lists are the mean and std of the tensors
             accordingly. see example in multiple_tensors_mse_loss
            log_function (Callable): Function to log information about the GPTQ process.
            train_bias (bool): Whether to update the bias during the training or not.
            rounding_type (RoundingType): An enum that defines the rounding type.
            use_hessian_based_weights (bool): Whether to use Hessian-based weights for weighted average loss.
            optimizer_quantization_parameter (Any): Optimizer to override the rest optimizer  for quantizer parameters.
            optimizer_bias (Any): Optimizer to override the rest optimizer for bias.
            regularization_factor (float): A floating point number that defines the regularization factor.
            hessian_weights_config (GPTQHessianWeightsConfig): A configuration that include all necessary arguments to run a computation of Hessian weights for the GPTQ loss.
            gptq_quantizer_params_override (dict): A dictionary of parameters to override in GPTQ quantizer instantiation. Defaults to None (no parameters).

        """
        self.n_iter = n_iter
        self.optimizer = optimizer
        self.optimizer_rest = optimizer_rest
        self.loss = loss
        self.log_function = log_function
        self.train_bias = train_bias

        self.rounding_type = rounding_type
        self.use_hessian_based_weights = use_hessian_based_weights
        self.optimizer_quantization_parameter = optimizer_quantization_parameter
        self.optimizer_bias = optimizer_bias
        self.regularization_factor = regularization_factor
        self.hessian_weights_config = hessian_weights_config

        self.gptq_quantizer_params_override = {} if gptq_quantizer_params_override is None \
            else gptq_quantizer_params_override


class GradientPTQConfigV2(GradientPTQConfig):
    """
    Configuration to use for quantization with GradientPTQV2 (experimental).
    """
    def __init__(self, n_epochs: int,
                 optimizer: Any,
                 optimizer_rest: Any = None,
                 loss: Callable = None,
                 log_function: Callable = None,
                 train_bias: bool = True,
                 rounding_type: RoundingType = RoundingType.SoftQuantizer,
                 use_hessian_based_weights: bool = True,
                 optimizer_quantization_parameter: Any = None,
                 optimizer_bias: Any = None,
                 regularization_factor: float = REG_DEFAULT,
                 hessian_weights_config: GPTQHessianWeightsConfig = GPTQHessianWeightsConfig(),
                 gptq_quantizer_params_override: Dict[str, Any] = None):
        """
        Initialize a GradientPTQConfigV2.

        Args:
            n_epochs (int): Number of representative dataset epochs to train.
            optimizer (Any): Optimizer to use.
            optimizer_rest (Any): Optimizer to use for bias and quantizer parameters.
            loss (Callable): The loss to use. should accept 6 lists of tensors. 1st list of quantized tensors, the 2nd list is the float tensors,
             the 3rd is a list of quantized weights, the 4th is a list of float weights, the 5th and 6th lists are the mean and std of the tensors
             accordingly. see example in multiple_tensors_mse_loss
            log_function (Callable): Function to log information about the GPTQ process.
            train_bias (bool): Whether to update the bias during the training or not.
            rounding_type (RoundingType): An enum that defines the rounding type.
            use_hessian_based_weights (bool): Whether to use Hessian-based weights for weighted average loss.
            optimizer_quantization_parameter (Any): Optimizer to override the rest optimizer  for quantizer parameters.
            optimizer_bias (Any): Optimizer to override the rest optimizerfor bias.
            regularization_factor (float): A floating point number that defines the regularization factor.
            hessian_weights_config (GPTQHessianWeightsConfig): A configuration that include all necessary arguments to run a computation of Hessian weights for the GPTQ loss.
            gptq_quantizer_params_override (dict): A dictionary of parameters to override in GPTQ quantizer instantiation. Defaults to None (no parameters).

        """

        super().__init__(n_iter=None,
                         optimizer=optimizer,
                         optimizer_rest=optimizer_rest,
                         loss=loss,
                         log_function=log_function,
                         train_bias=train_bias,
                         rounding_type=rounding_type,
                         use_hessian_based_weights=use_hessian_based_weights,
                         optimizer_quantization_parameter=optimizer_quantization_parameter,
                         optimizer_bias=optimizer_bias,
                         regularization_factor=regularization_factor,
                         hessian_weights_config=hessian_weights_config,
                         gptq_quantizer_params_override=gptq_quantizer_params_override)
        self.n_epochs = n_epochs

    @classmethod
    def from_v1(cls, n_ptq_iter: int, config_v1: GradientPTQConfig):
        """
        Initialize a GradientPTQConfigV2 from GradientPTQConfig instance.

        Args:
            n_ptq_iter (int): Number of PTQ calibration iters (length of representative dataset).
            config_v1 (GradientPTQConfig): A GPTQ config to convert to V2.

        """
        n_epochs = int(round(config_v1.n_iter) / n_ptq_iter)
        v1_params = config_v1.__dict__
        v1_params = {k: v for k, v in v1_params.items() if k != 'n_iter'}
        return cls(n_epochs, **v1_params)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple, List
from matplotlib import pyplot as plt
from matplotlib.figure import Figure

from model_compression_toolkit.core.common import Graph, BaseNode


def get_kernel_layer_represent_name(node: BaseNode) -> str:
    """
    Returns the mapping between a layer's type and its name to appear in the visualization figure.
    We apply this function only to map types of layers that have kernels (other layers do not appear by name
    in the figures)

    Args:
        node: A graph node representing a model's layer.

    Returns: The name of the layer to appear in the visualization.

    """

    return node.type.__name__


class WeightsFinalBitwidthConfigVisualizer:
    """
    Class to visualize the chosen bit-width configuration for weights configurable layers in mixed-precision mode.
    WeightsFinalBitwidthConfigVisualizer draws a bar plot with the bit-width value of each layer.
    """
    def __init__(self,
                 final_weights_nodes_config: List[Tuple[BaseNode, int]]):
        """
        Initialize a WeightsFinalBitwidthConfigVisualizer object.
        Args:
            final_weights_nodes_config: List of candidates' indices of sorted weights configurable nodes
                (expects a list of tuples - (node, node's final weights bitwidth).
        """

        self.final_weights_nodes_config = final_weights_nodes_config
        self.node_reps_names = [get_kernel_layer_represent_name(node_cfg[0]) for node_cfg in self.final_weights_nodes_config]
        self.node_final_bitwidth = [node_cfg[1] for node_cfg in self.final_weights_nodes_config]
        self.bitwidth_colors_map = {2: 'tomato', 4: 'royalblue', 8: 'limegreen'}
        self.configs_colors = [self.bitwidth_colors_map[b] for b in self.node_final_bitwidth]
        self.bar_width = 2
        self.gap = self.bar_width + 3

    def plot_config_bitwidth(self) -> Figure:
        """
        Plots a bar figure with the layers' bit-width values according to the chosen config.

        Returns:
            Figure of the layers' bit-width values.
        """

        layers_loc = [i for i in range(self.bar_width, self.bar_width + self.gap * len(self.node_reps_names), self.gap)]
        fig, ax = plt.subplots(figsize=(20, 10))
        plt.bar(layers_loc, self.node_final_bitwidth, color=self.configs_colors, width=self.bar_width, align='center')
        plt.xticks(layers_loc, self.node_reps_names, rotation='vertical')
        plt.rc('xtick', labelsize=4)
        plt.rc('ytick', labelsize=4)
        plt.xlabel('Layers', fontsize=12)
        plt.ylabel('Number of bits', fontsize=12)
        plt.tight_layout()
        return fig


class ActivationFinalBitwidthConfigVisualizer:
    """
    Class to visualize the activation configuration attributes.
    ActivationFinalBitwidthConfigVisualizer can draw a figure of the chosen bit-width configuration for activation
    configurable layers in mixed-precision mode.
    It also allows to draw a figure with the activation tensors memory size.
    """

    def __init__(self,
                 final_activation_nodes_config: List[Tuple[BaseNode, int]]):
        """
        Initialize a ActivationFinalBitwidthConfigVisualizer object.
        Args:
            final_activation_nodes_config: List of candidates' indices of sorted activation configurable nodes.
                (expects a list of tuples - (node, node's final activation bitwidth).
        """

        self.final_activation_nodes_config = final_activation_nodes_config
        self.node_final_bitwidth = [node_cfg[1] for node_cfg in self.final_activation_nodes_config]
        self.bar_width = 1
        self.vis_comp_rates = {4.0: 'tomato', 8.0: 'orange', 12.0: 'limegreen'}

    def plot_config_bitwidth(self) -> Figure:
        """
        Plots a bar figure with the layers' bit-width values according to the chosen config.

        Returns:
            Figure of the layers' bit-width values.
        """

        layers_loc = [i for i in range(1, len(self.node_final_bitwidth) + 1)]
        fig, ax = plt.subplots()
        plt.bar(layers_loc, self.node_final_bitwidth, width=self.bar_width, align='center')
        plt.grid()
        plt.xlabel('Layer Index', fontsize=12)
        plt.ylabel('Number of bits', fontsize=12)
        plt.tight_layout()
        return fig

    def plot_tensor_sizes(self, graph: Graph) -> Figure:
        """
        Plots a bar figure with the layers' activation tensors sizes.
        Also, adds horizontal line indicators for the max tensor size (in MB) for the defined set of compression rates.

        Args:
            graph: A Graph object to be used for calculating the layers' activation tensor sizes.

        Returns:
            Figure of the layers' activation tensors sizes.
        """

        tensors_sizes = [4.0 * n.get_total_output_params() / 1000000.0
                         for n in graph.get_sorted_activation_configurable_nodes()]  # in MB
        max_tensor_size = max(tensors_sizes)
        max_lines = [(rate, max_tensor_size / rate, color) for rate, color in self.vis_comp_rates.items()]

        layers_loc = [i for i in range(1, len(self.final_activation_nodes_config) + 1)]
        fig, ax = plt.subplots()
        plt.bar(layers_loc, tensors_sizes, width=self.bar_width, align='center')
        plt.grid()
        plt.xlabel('Layer Index', fontsize=12)
        plt.ylabel('Tensor Size [MB]', fontsize=12)

        # add max tensor lines
        for rate, t_size, color in max_lines:
            plt.plot([layers_loc[0], layers_loc[-1]], [t_size, t_size], "k--", color=color, label=f"ACR = {rate}")

        plt.legend()
        plt.tight_layout()
        return fig



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch
import torch.nn as nn
from typing import List
from model_compression_toolkit.core.pytorch.constants import BIAS
from model_compression_toolkit.core.pytorch.default_framework_info import DEFAULT_PYTORCH_INFO
from model_compression_toolkit.gptq.common.gptq_graph import get_kernel_attribute_name_for_gptq
from model_compression_toolkit.quantizers_infrastructure import PytorchQuantizationWrapper
from model_compression_toolkit.quantizers_infrastructure.trainable_infrastructure.common.base_trainable_quantizer import VariableGroup


def get_gptq_trainable_parameters(fxp_model: nn.Module,
                                  add_bias: bool = False,
                                  ) -> (List[nn.Parameter], List[nn.Parameter], List[nn.Parameter]):
    """
    Get trainable parameters from all layers in a model

    Args:
        fxp_model: Model to get its trainable parameters.
        add_bias: Whether to include biases of the model (if there are) or not.

    Returns:
        A list of trainable variables in a model. Each item is a list of a layers weights.
    """

    trainable_aux_weights = nn.ParameterList()
    trainable_threshold = nn.ParameterList()
    trainable_bias = nn.ParameterList()

    for layer in fxp_model.modules():
        if isinstance(layer, PytorchQuantizationWrapper):
            kernel_attribute = get_kernel_attribute_name_for_gptq(layer_type=type(layer.layer),
                                                                  fw_info=DEFAULT_PYTORCH_INFO)

            # collect trainable weights per quantizer
            quantizer_trainable_weights = layer.weights_quantizers[kernel_attribute].get_trainable_variables(VariableGroup.WEIGHTS)
            quantizer_trainable_threshold = layer.weights_quantizers[kernel_attribute].get_trainable_variables(VariableGroup.QPARAMS)
            trainable_aux_weights.extend(quantizer_trainable_weights)
            trainable_threshold.extend(quantizer_trainable_threshold)

            if add_bias and hasattr(layer.layer, BIAS):
                bias = getattr(layer.layer, BIAS)
                trainable_bias.append(bias)

    return trainable_aux_weights, trainable_bias, trainable_threshold


def get_weights_for_loss(fxp_model: nn.Module) -> [List[nn.Parameter], List[torch.Tensor]]:
    """
    Get all float and quantized kernels for the GPTQ loss

    Args:
        fxp_model: Model to get its float and quantized weights.

    Returns:
        A list of float kernels, each item is the float kernel of the layer
        A list of quantized kernels, each item is the quantized kernel of the layer
    """

    flp_weights_list, fxp_weights_list = [], []
    for layer in fxp_model.modules():
        if isinstance(layer, PytorchQuantizationWrapper):
            # Collect pairs of float and quantized weights per layer
            for weight, quantizer_vars, quantizer in layer.get_weights_vars():
                flp_weights_list.append(quantizer_vars)
                fxp_weights_list.append(quantizer(training=False, inputs=quantizer_vars))

    return flp_weights_list, fxp_weights_list



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import warnings
from typing import List

import numpy as np

from model_compression_toolkit.core.common.logger import Logger
from model_compression_toolkit.core.common.constants import FOUND_TF

from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.quantizers_infrastructure import QuantizationTarget
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import mark_quantizer
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.constants import MULTIPLIER_N_BITS, EPS

if FOUND_TF:
    import tensorflow as tf
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.keras.quantizers.base_keras_inferable_quantizer \
        import BaseKerasInferableQuantizer
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.keras.quantizer_utils import lut_quantizer


    @mark_quantizer(quantization_target=QuantizationTarget.Activation,
                    quantization_method=[QuantizationMethod.LUT_POT_QUANTIZER],
                    quantizer_type=None)
    class ActivationLutPOTInferableQuantizer(BaseKerasInferableQuantizer):
        """
        Class for quantizing activations using lut power-of-two quantizer
        """

        def __init__(self,
                     num_bits: int,
                     cluster_centers: np.ndarray,
                     threshold: List[float],
                     signed: bool,
                     multiplier_n_bits: int = MULTIPLIER_N_BITS,
                     eps: float = EPS):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                cluster_centers: the cluster centers to assign the activations
                threshold: threshold for quantizing activations
                signed: whether or not to use signed quantization
                multiplier_n_bits: Number of bits that determines the quantization range
                eps: Small value for numerical stability in division
            """
            # Call the superclass constructor with the given parameters, along with the target of Activation
            # quantization
            super(ActivationLutPOTInferableQuantizer, self).__init__()

            assert isinstance(threshold, list), f'Expected threshold to be of type list but is {type(threshold)}'
            assert all([isinstance(x, (float, np.float32, tf.float32)) for x in
                        threshold]), f'Expected threshold list to contain float or np.float values but found ' \
                                     f'{[type(x) for x in threshold]}'

            # In activation per-channel quantization is not supported thus we expect a single threshold value.
            assert len(threshold) == 1, f'In per-tensor quantization threshold should be of ' \
                                        f'length 1 but is {len(threshold)}'

            is_threshold_pot = np.all([int(np.log2(x)) == np.log2(x) for x in threshold])
            assert is_threshold_pot, f'Expected threshold to be power of 2 but is {threshold}'

            self.threshold = threshold[0]

            assert len(np.unique(cluster_centers)) <= 2 ** num_bits, \
                f'Expected num of cluster centers to be less or equal than {2 ** num_bits} ' \
                f'but got {len(cluster_centers)}'

            assert not np.any(cluster_centers - cluster_centers.astype(int)), f'Expected cluster centers to be integers'

            if signed:
                assert np.all((-1 * (2 ** (multiplier_n_bits - int(signed))) <= cluster_centers) &
                              (cluster_centers <= (2 ** (multiplier_n_bits - int(signed)) - 1))), \
                    f'Expected cluster centers in the quantization range'
            else:
                assert np.all(cluster_centers <= (2 ** multiplier_n_bits)), \
                    f'Expected cluster centers in the quantization range'

            # num_bits must be less than multiplier_n_bits
            assert num_bits <= multiplier_n_bits, f'Look-Up-Table bit configuration has {num_bits} bits. It must be ' \
                                                  f'less then {multiplier_n_bits}'
            if num_bits == multiplier_n_bits:
                warnings.warn("Num of bits equal to multiplier n bits, Please be aware LUT quantizier may be "
                              "inefficient in that case, consider using SymmetricInferableQuantizer instead")

            # If unsigned activation quantization, all cluster_centers must have the same sign
            if not signed:
                assert np.all(cluster_centers >= 0), f'Expected unsigned cluster centers in unsigned activation ' \
                                                     f'quantization '

            self.num_bits = num_bits
            self.cluster_centers = cluster_centers
            self.signed = signed
            self.multiplier_n_bits = multiplier_n_bits
            self.eps = eps

        def __call__(self, inputs: tf.Tensor) -> tf.Tensor:
            """
            Quantize the given inputs using the quantizer parameters.

            Args:
                inputs: input tensor to quantize

            Returns:
                quantized tensor.
            """
            assert inputs.dtype == tf.float32, f'Input tensor was expected to be a float tensor but is of type ' \
                                               f'{inputs.dtype}'

            return lut_quantizer(inputs, cluster_centers=self.cluster_centers, signed=self.signed,
                                 threshold=self.threshold, multiplier_n_bits=self.multiplier_n_bits, eps=self.eps)

        def get_config(self):
            """
            Return a dictionary with the configuration of the quantizer.

            Returns:
                Dictionary with the following keys: 'num_bits', 'cluster_centers', 'threshold', 'signed',
                'multiplier_n_bits', 'eps'
            """
            return {'num_bits': self.num_bits,
                    'cluster_centers': self.cluster_centers,
                    'threshold': self.threshold,
                    'signed': self.signed,
                    'multiplier_n_bits': self.multiplier_n_bits,
                    'eps': self.eps}


else:
    class ActivationLutPOTInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            Logger.error('Installing tensorflow and tensorflow_model_optimization is mandatory '
                         'when using ActivationLutPOTInferableQuantizer. '
                         'Could not find Tensorflow package.')



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import numpy as np

from model_compression_toolkit.core.common.logger import Logger
from model_compression_toolkit.core.common.constants import FOUND_TF
from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.quantizers_infrastructure import QuantizationTarget
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import mark_quantizer
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.quant_utils import \
    adjust_range_to_include_zero
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.keras.validation_functions import \
    validate_uniform_min_max_ranges, validate_adjusted_min_max_ranges

if FOUND_TF:
    import tensorflow as tf
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.keras.quantizers.base_keras_inferable_quantizer import BaseKerasInferableQuantizer

    @mark_quantizer(quantization_target=QuantizationTarget.Activation,
                    quantization_method=[QuantizationMethod.UNIFORM],
                    quantizer_type=None)
    class ActivationUniformInferableQuantizer(BaseKerasInferableQuantizer):
        """
        Class for quantizing activations using an uniform quantizer
        """

        def __init__(self,
                     num_bits: int,
                     min_range: List[float],
                     max_range: List[float],
                     ):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                min_range: min range for quantizing activations
                max_range: max range for quantizing activations
            """
            super(ActivationUniformInferableQuantizer, self).__init__()

            # Validate ranges properties
            validate_uniform_min_max_ranges(min_range,
                                            max_range)

            # In activation per-channel quantization is not supported thus we expect a single min/max value.
            assert len(min_range) == 1, f'In per-tensor quantization min_range should be of length 1 but is {len(min_range)}'
            assert len(max_range) == 1, f'In per-tensor quantization max_range should be of length 1 but is {len(max_range)}'

            self.num_bits = num_bits

            # Convert min/max to numpy arrays
            min_range, max_range = np.asarray(min_range), np.asarray(max_range)
            _min_range, _max_range = adjust_range_to_include_zero(min_range, max_range, num_bits)
            validate_adjusted_min_max_ranges(min_range=min_range,
                                             max_range=max_range,
                                             adj_min=_min_range,
                                             adj_max=_max_range)

            self.max_range = _max_range[0]
            self.min_range = _min_range[0]


        def __call__(self, inputs:tf.Tensor) -> tf.Tensor:
            """
            Quantize the given inputs using the quantizer parameters.

            Args:
                inputs: input tensor to quantize

            Returns:
                quantized tensor.
            """
            assert inputs.dtype==tf.float32, f'Input tensor was expected to be a float tensor but is of type {inputs.dtype}'

            return tf.quantization.fake_quant_with_min_max_vars(inputs,
                                                                min=self.min_range,
                                                                max=self.max_range,
                                                                num_bits=self.num_bits)

        def get_config(self):
            """
            Return a dictionary with the configuration of the quantizer.

            Returns:
                Dictionary with the following keys: 'num_bits', 'min_range', 'max_range'
            """
            return {'num_bits': self.num_bits,
                    'min_range': self.min_range,
                    'max_range': self.max_range}

else:
    class ActivationUniformInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            Logger.error('Installing tensorflow and tensorflow_model_optimization is mandatory '
                         'when using ActivationUniformInferableQuantizer. '
                         'Could not find Tensorflow package.')



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple

import torch
import numpy as np


def get_working_device():
    """
    Get the working device of the environment

    Returns:
        Device "cuda" if GPU is available, else "cpu"

    """
    return torch.device('cuda' if torch.cuda.is_available() else 'cpu')


def to_torch_tensor(tensor):
    """
    Convert a Numpy array to a Torch tensor.
    Args:
        tensor: Numpy array.

    Returns:
        Torch tensor converted from the input Numpy array.
    """
    working_device = get_working_device()
    if isinstance(tensor, torch.Tensor):
        return tensor.to(working_device)
    elif isinstance(tensor, list):
        return [to_torch_tensor(t) for t in tensor]
    elif isinstance(tensor, tuple):
        return (to_torch_tensor(t) for t in tensor)
    elif isinstance(tensor, np.ndarray):
        return torch.from_numpy(tensor.astype(np.float32)).to(working_device)
    elif isinstance(tensor, float):
        return torch.Tensor([tensor]).to(working_device)
    elif isinstance(tensor, int):
        return torch.Tensor([tensor]).int().to(working_device)
    else:
        raise Exception(f'Conversion of type {type(tensor)} to {type(torch.Tensor)} is not supported')


def fix_range_to_include_zero(range_min: torch.Tensor,
                              range_max: torch.Tensor,
                              n_bits: int) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Adjusting the quantization range to include representation of 0.0 in the quantization grid.
    If quantization per-channel, then range_min and range_max should be tensors in the specific shape that allows
    quantization along the channel_axis.
    Args:
        range_min: min bound of the quantization range (before adjustment).
        range_max: max bound of the quantization range (before adjustment).
        n_bits: Number of bits to quantize the tensor.
    Returns: adjusted quantization range
    """
    min_positive = range_min > 0
    max_negative = range_max < 0
    mid_range = torch.logical_and(torch.logical_not(min_positive), torch.logical_not(max_negative))
    min_positive = min_positive.float()
    max_negative = max_negative.float()
    mid_range = mid_range.float()

    scale = (range_max - range_min) / (2 ** n_bits - 1)
    min_range_adj = scale * torch.round(range_min / scale)
    max_range_adj = range_max - range_min + min_range_adj

    min_range_adj = min_range_adj * mid_range + max_negative * range_min
    max_range_adj = max_range_adj * mid_range + min_positive * range_max
    return min_range_adj, max_range_adj


def lut_quantizer(tensor_data: torch.Tensor,
                  cluster_centers: torch.Tensor,
                  signed: bool,
                  threshold: torch.Tensor,
                  multiplier_n_bits: int,
                  eps: float) -> torch.Tensor:
    """
    Quantize a tensor using a non-uniform quantization based on the pre-defined clusters.
    1. Scales tensor_data with the threshold into n-bit quantization range.
    2. Assigns cluster centers to each value.
    3. Scales back by multiplying the result by threshold and dividing with the quantization range max value.
    The result is the quantized tensor.

    Args:
        tensor_data: Input activation tensor.
        cluster_centers: The cluster centers to assign the tensor values.
        signed: Whether the quantization is signed or not.
        threshold: Threshold for quantization.
        multiplier_n_bits: Number of bits that determines the quantization range
        eps: Small value for numerical stability in division.

    Returns: Quantized tensor.
    """

    tensor = int_quantization_with_threshold(tensor_data, n_bits=multiplier_n_bits, signed=signed, threshold=threshold,
                                             eps=eps)
    tensor = tensor.unsqueeze(-1)

    expanded_cluster_centers = cluster_centers.reshape([*[1 for _ in range(len(tensor.shape) - 1)], -1])
    cluster_assignments = torch.argmin(torch.abs(tensor - expanded_cluster_centers), dim=-1)
    centers = cluster_centers.flatten()[cluster_assignments]

    quant_tensor = (centers / (2 ** (multiplier_n_bits - int(signed)))) * threshold

    return quant_tensor


def int_quantization_with_threshold(data: torch.Tensor,
                                    n_bits: int,
                                    signed: bool,
                                    threshold: torch.Tensor,
                                    eps: float) -> torch.Tensor:
    """
    Divides data by threshold and quantize it to integers in the quantization range (depends on signed value).

    Args:
        data: Tensor data.
        n_bits: Number of bits that determines the quantization range.
        signed: Whether the quantization is signed or not.
        threshold: Threshold for quantization.
        eps: Small value for numerical stability in division.

    Returns:
        Uniform Quantized tensor.

    """

    if signed:
        clip_max = 2 ** (n_bits - 1) - 1
        clip_min = -2 ** (n_bits - 1)
    else:
        clip_max = 2 ** n_bits - 1
        clip_min = 0

    return torch.clip((data / (threshold + eps)) * (2 ** (n_bits - int(signed))),
                      min=clip_min, max=clip_max)



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import tensorflow as tf
import numpy as np

from model_compression_toolkit.gptq import RoundingType
from model_compression_toolkit import quantizers_infrastructure as qi
from model_compression_toolkit.core.common import max_power_of_two
from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.gptq.common.gptq_constants import PTQ_THRESHOLD, SCALE_PTQ, \
    SOFT_ROUNDING_GAMMA, SOFT_ROUNDING_ZETA, AUXVAR
from model_compression_toolkit.gptq.keras.quantizer import quant_utils as qutils
from typing import Dict, Any
from model_compression_toolkit.core.common.constants import THRESHOLD, MIN_THRESHOLD
from model_compression_toolkit.gptq.keras.quantizer.base_keras_gptq_quantizer import BaseKerasGPTQTrainableQuantizer
from model_compression_toolkit.gptq.keras.quantizer.quant_utils import power_of_two_max, clip, calculate_delta
from model_compression_toolkit.quantizers_infrastructure import TrainableQuantizerWeightsConfig
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import mark_quantizer
from model_compression_toolkit.quantizers_infrastructure.trainable_infrastructure.common.quant_utils import \
    get_threshold_reshape_shape
from model_compression_toolkit.quantizers_infrastructure.trainable_infrastructure.common.base_trainable_quantizer import VariableGroup


def soft_rounding_symmetric_quantizer(input_tensor: tf.Tensor,
                                      auxvar_tensor: tf.Variable,
                                      threshold_tensor: tf.Tensor,
                                      num_bits: int,
                                      signed: bool,
                                      power_of_two: bool) -> tf.Tensor:
    """
    Quantize a tensor symmetrically for GPTQ quantizers.

    Args:
        input_tensor: Tensor to quantize. values of this tensor are not changed during gptq.
        auxvar_tensor: Tensor that manifests the bit shift of the quantized weights due to gptq training.
        threshold_tensor: Tensor with values to compute the threshold.
        num_bits: Num of bits to use.
        signed: Signedness of the quantization range.
        power_of_two: Whether the threshold should be constrained or not.

    Returns:
        A quantized tensor.
    """

    if power_of_two:
        threshold_tensor = power_of_two_max(threshold_tensor)
    delta = calculate_delta(threshold_tensor, num_bits, signed)
    input_tensor = tf.stop_gradient(input_tensor)
    input_tensor_int = tf.floor(input_tensor / delta)
    tensor_q = input_tensor_int + auxvar_tensor
    min_int = -int(signed) * (2 ** (num_bits - int(signed)))
    max_int = (2 ** (num_bits - int(signed))) - 1
    return delta * clip(tensor_q, max_val=max_int, min_val=min_int)


@mark_quantizer(quantization_target=qi.QuantizationTarget.Weights,
                quantization_method=[QuantizationMethod.POWER_OF_TWO, QuantizationMethod.SYMMETRIC],
                quantizer_type=RoundingType.SoftQuantizer)
class SymmetricSoftRoundingGPTQ(BaseKerasGPTQTrainableQuantizer):
    """
    Trainable symmetric quantizer to optimize the rounding of the quantized values using a soft quantization method.
    """

    def __init__(self,
                 quantization_config: TrainableQuantizerWeightsConfig,
                 quantization_parameter_learning: bool = False):
        """
        Initialize a SymmetricSoftRoundingGPTQ object with parameters to use
        for the quantization.

        Args:
            quantization_config: Trainable weights quantizer config.
            quantization_parameter_learning: Whether to train the quantization threshold.
        """
        super().__init__(quantization_config)
        self.num_bits = quantization_config.weights_n_bits
        self.per_channel = quantization_config.weights_per_channel_threshold

        threshold_values = quantization_config.weights_quantization_params[THRESHOLD]
        self.threshold_shape = np.asarray(threshold_values).shape
        self.threshold_values = np.reshape(np.asarray(threshold_values), [-1]) if self.per_channel else np.asarray(
            threshold_values)

        self.quantization_axis = quantization_config.weights_channels_axis
        self.power_of_two = quantization_config.weights_quantization_method == QuantizationMethod.POWER_OF_TWO
        self.quantization_parameter_learning = quantization_parameter_learning
        self.num_channels = len(self.threshold_values) if self.per_channel else 1

        # gamma and zeta are stretch parameters for computing the rectified sigmoind function.
        # See: https://arxiv.org/pdf/2004.10568.pdf
        self.gamma = SOFT_ROUNDING_GAMMA
        self.zeta = SOFT_ROUNDING_ZETA

        self.quantizer_parameters = {}

    def initialize_quantization(self,
                                tensor_shape: Any,
                                name: str,
                                layer: Any):
        """
        Add quantizer parameters to the quantizer parameters dictionary

        Args:
            tensor_shape: tensor shape of the quantized tensor.
            name: Tensor name.
            layer: Layer to quantize.
        """

        if self.per_channel:
            reshape_shape = get_threshold_reshape_shape(tensor_shape,
                                                        quant_axis=self.quantization_axis,
                                                        quant_axis_dim=self.num_channels)
        else:
            reshape_shape = [self.num_channels]

        ptq_threshold_tensor = layer.add_weight(
            f"{name}_{PTQ_THRESHOLD}",
            shape=reshape_shape,
            initializer=tf.keras.initializers.Constant(1.0),
            trainable=False)
        ptq_threshold_tensor.assign(self.threshold_values.reshape(reshape_shape))

        w = getattr(layer.layer, name)
        auxvar_tensor = layer.add_weight(
            f"{name}_{AUXVAR}",
            shape=list(w.shape),
            initializer=tf.keras.initializers.Constant(0.0),
            trainable=True)

        delta = qutils.calculate_delta(ptq_threshold_tensor, self.num_bits, signed=True)
        w_floor = tf.floor(w / delta)
        rest = (w / delta) - w_floor  # rest of rounding [0, 1)
        # Note that (rest - self.gamma) can't be zero since rest is positive and gamma is negative, so the division
        # is safe
        alpha = -qutils.safe_log((self.zeta - self.gamma) / (rest - self.gamma) - 1, 1e-16)  # => sigmoid(alpha) = rest

        auxvar_tensor.assign(alpha)

        # Add quantization variables
        self.add_quantizer_variable(AUXVAR, auxvar_tensor, VariableGroup.WEIGHTS)
        self.add_quantizer_variable(PTQ_THRESHOLD, ptq_threshold_tensor, VariableGroup.QPARAMS)

        if self.quantization_parameter_learning and not self.power_of_two:
            scale = layer.add_weight(
                f"{name}_{SCALE_PTQ}",
                shape=self.num_channels,
                initializer=tf.keras.initializers.Constant(1.0),
                trainable=True)
            self.add_quantizer_variable(SCALE_PTQ, scale, VariableGroup.QPARAMS)

    def get_soft_targets(self) -> tf.Tensor:
        """
        Computes the rectified sigmoid function for the quantization target parameters.

        Returns:
            A tensor with the soft rounding targets values.

        """
        return qutils.clip(
            tf.sigmoid(self.get_quantizer_variable(AUXVAR)) * (self.zeta - self.gamma) + self.gamma, 1, 0)

    def __call__(self,
                 inputs: tf.Tensor,
                 training: bool):
        """
        Quantize a tensor.

        Args:
            inputs: Input tensor to quantize.
            training: Whether the graph is in training mode.

        Returns:
            The quantized tensor.
        """

        ptq_threshold_tensor = self.get_quantizer_variable(PTQ_THRESHOLD)

        #####################################################
        # Soft Rounding
        #####################################################
        aux_var = self.get_soft_targets()
        if not training:
            aux_var = tf.cast(tf.math.greater_equal(aux_var, 0.5), tf.float32)

        if self.per_channel:
            reshape_shape = get_threshold_reshape_shape(inputs.shape,
                                                        quant_axis=self.quantization_axis,
                                                        quant_axis_dim=-1)

            ##########################################################
            # Calculate soft rounding targets and optimized threshold
            ##########################################################
            ptq_threshold_tensor_hat = tf.reshape(ptq_threshold_tensor, reshape_shape)

            #####################################################
            # Quantized Input
            #####################################################
            q_tensor = soft_rounding_symmetric_quantizer(input_tensor=inputs,
                                                         auxvar_tensor=aux_var,
                                                         threshold_tensor=ptq_threshold_tensor_hat,
                                                         num_bits=self.num_bits,
                                                         signed=True,
                                                         power_of_two=self.power_of_two)

            if self.quantization_parameter_learning and not self.power_of_two:
                scale = tf.reshape(self.get_quantizer_variable(SCALE_PTQ), reshape_shape)
                q_tensor *= scale

        else:
            q_tensor = soft_rounding_symmetric_quantizer(input_tensor=inputs,
                                                         auxvar_tensor=aux_var,
                                                         threshold_tensor=ptq_threshold_tensor.value(),
                                                         num_bits=self.num_bits,
                                                         signed=True,
                                                         power_of_two=self.power_of_two)

            if self.quantization_parameter_learning and not self.power_of_two:
                scale = self.get_quantizer_variable(SCALE_PTQ)
                q_tensor *= scale

        return q_tensor

    def get_quant_config(self) -> Dict[str, np.ndarray]:
        """
        Returns the config used to edit NodeQuantizationConfig after GPTQ retraining

        Returns:
            A dictionary of attributes the quantize_config retraining has changed during GPTQ retraining.
            Keys must match NodeQuantizationConfig attributes
        """

        if self.power_of_two:
            old_threshold = self.get_quantizer_variable(PTQ_THRESHOLD)
            old_threshold = max_power_of_two(old_threshold, MIN_THRESHOLD)

        else:
            old_threshold = self.get_quantizer_variable(PTQ_THRESHOLD)
            if self.quantization_parameter_learning:
                scale = tf.reshape(self.get_quantizer_variable(SCALE_PTQ), self.threshold_shape)
                old_threshold = old_threshold * scale
            old_threshold = old_threshold.numpy()
        old_threshold = old_threshold.reshape(self.threshold_shape)
        return {THRESHOLD: old_threshold}



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import numpy as np

from model_compression_toolkit.core.common.constants import FOUND_TF
from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import mark_quantizer, QuantizationTarget
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.quant_utils import \
    adjust_range_to_include_zero
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.keras.validation_functions import \
    validate_uniform_min_max_ranges, validate_adjusted_min_max_ranges

if FOUND_TF:
    import tensorflow as tf
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.keras.quantizers.base_keras_inferable_quantizer import BaseKerasInferableQuantizer

    @mark_quantizer(quantization_target=QuantizationTarget.Weights,
                    quantization_method=[QuantizationMethod.UNIFORM],
                    quantizer_type=None)
    class WeightsUniformInferableQuantizer(BaseKerasInferableQuantizer):
        """
        Class for quantizing weights using a uniform quantizer
        """
        def __init__(self,
                     num_bits: int,
                     min_range: List[float],
                     max_range: List[float],
                     per_channel: bool,
                     channel_axis: int = None,
                     input_rank: int = None
                     ):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                min_range: min quantization range for quantizing weights
                max_range: max quantization range for quantizing weights
                per_channel: whether to use per-channel quantization
                channel_axis: axis along which to apply per-channel quantization
                input_rank: number of dimensions of input tensor the quantizer quantizes
            """

            super(WeightsUniformInferableQuantizer, self).__init__()

            # Validate inputs properties
            validate_uniform_min_max_ranges(min_range,
                                            max_range)

            # Convert min/max to numpy arrays
            min_range, max_range = np.asarray(min_range), np.asarray(max_range)
            _min_range, _max_range = adjust_range_to_include_zero(min_range, max_range, num_bits)
            validate_adjusted_min_max_ranges(min_range=min_range,
                                             max_range=max_range,
                                             adj_min=_min_range,
                                             adj_max=_max_range)

            self.num_bits = num_bits
            self.max_range = _max_range
            self.min_range = _min_range

            if per_channel:
                assert input_rank is not None, f'Input rank is missing in per channel quantization'
                assert channel_axis is not None, f'Channel axis is missing in per channel quantization'
                assert len(self.min_range) >= 1, f'In per-channel quantization min ranges list should be of length >= 1 but is {len(self.min_range)}'
                assert len(self.max_range) >= 1, f'In per-channel quantization max ranges list should be of length >= 1 but is {len(self.max_range)}'
            else:
                assert len(self.min_range) == 1, f'In per-tensor quantization min/max should be of length 1 but is {len(min_range)}'
                assert len(self.min_range) == 1, f'In per-tensor quantization min_range should be of length 1 but is {len(self.min_range)}'
                assert len(self.max_range) == 1, f'In per-tensor quantization max_range should be of length 1 but is {len(self.max_range)}'
                self.min_range = self.min_range[0]
                self.max_range = self.max_range[0]

            self.per_channel = per_channel
            self.channel_axis = channel_axis
            self.input_rank = input_rank

            # Tensorflow's fake_quant_with_min_max_vars_per_channel only works on last axis, so
            # need to move the quantization axis to the last axis
            if per_channel and channel_axis not in [-1, self.input_rank - 1]:
                # If per-channel quantization is being used and the channel axis is not the last axis,
                # create a permutation vector to move the channel axis to the last position
                self.perm_vec = list(np.arange(self.input_rank))
                self.perm_vec[channel_axis] = self.input_rank - 1
                self.perm_vec[self.input_rank - 1] = channel_axis
            else:
                # If per-channel quantization is not being used or the channel axis is already the last axis,
                # set the permutation vector to None
                self.perm_vec = None

        def __call__(self, inputs: tf.Tensor) -> tf.Tensor:
            """
            Quantize the given inputs using the quantizer parameters.

            Args:
                inputs: input tensor to quantize

            Returns:
                quantized tensor.
            """
            assert inputs.dtype==tf.float32, f'Input tensor was expected to be a float tensor but is of type {inputs.dtype}'

            # If per-channel quantization is being used
            if self.per_channel:
                # If a permutation vector has been created to move the channel axis to the last position
                if self.perm_vec:
                    # Transpose the input tensor to move the channel axis to the last position
                    inputs = tf.transpose(inputs, perm=self.perm_vec)

                # Quantize the input tensor using per-channel quantization
                q_tensor = tf.quantization.fake_quant_with_min_max_vars_per_channel(inputs,
                                                                                    min=self.min_range,
                                                                                    max=self.max_range,
                                                                                    num_bits=self.num_bits)
                if self.perm_vec:
                    # Transpose the quantized tensor back to its original shape
                    q_tensor = tf.transpose(q_tensor, perm=self.perm_vec)

                # Return the quantized tensor
                return q_tensor
            else:
                # If per-channel quantization is not being used, quantize the input tensor using regular quantization
                return tf.quantization.fake_quant_with_min_max_vars(inputs,
                                                                    min=self.min_range,
                                                                    max=self.max_range,
                                                                    num_bits=self.num_bits)


        def get_config(self):
            """
            Return a dictionary with the configuration of the quantizer.

            Returns:
                Dictionary with the following keys: 'num_bits', 'min_range', 'max_range', 'per_channel', 'channel_axis'
            """
            return {'per_channel': self.per_channel,
                    'num_bits': self.num_bits,
                    'max_range': self.max_range,
                    'min_range': self.min_range,
                    'channel_axis': self.channel_axis,
                    'input_rank': self.input_rank}


else:
    class WeightsUniformInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing tensorflow and tensorflow_model_optimization is mandatory '
                            'when using WeightsUniformInferableQuantizer. '
                            'Could not find Tensorflow package.')



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any

import numpy as np

from model_compression_toolkit.core.common import Logger


def validate_uniform_min_max_ranges(min_range: Any, max_range: Any) -> None:
    """
    Validate min/max ranges in uniform quantizers are valid

    Args:
        min_range: min range list to check
        max_range: max range list to check

    """
    assert isinstance(min_range, list), f'Expected min_range to be of type list but is {type(min_range)}'
    assert isinstance(max_range, list), f'Expected max_range to be of type list but is {type(max_range)}'

    assert all([isinstance(x, (float, np.float32, np.float64)) for x in
                min_range]), f'Expected min_range list to contain float values but found {[type(x) for x in min_range]}'
    assert all([isinstance(x, (float, np.float32, np.float64)) for x in
                max_range]), f'Expected max_range list to contain float values but found {[type(x) for x in max_range]}'

    assert len(min_range) == len(
        max_range), f'Expected min/max values to have the same length but min shape: {len(min_range)} and max shape: ' \
                    f'{len(max_range)}'

    # Convert min/max to numpy arrays
    min_range, max_range = np.asarray(min_range), np.asarray(max_range)
    assert np.all(max_range > min_range), f'Expected max_range to be bigger than min_range!'


def validate_adjusted_min_max_ranges(min_range: Any,
                                     max_range: Any,
                                     adj_min:Any,
                                     adj_max:Any) -> None:
    """
    Validate adjusted min/max ranges in uniform quantization are valid

    Args:
        min_range: original min range
        max_range: original max range
        adj_min: adjusted min range
        adj_max: adjusted max range

    """

    assert np.all(adj_min <= 0) and np.all(
        adj_max >= 0), f'Expected zero to be in the range, got min_range={adj_min}, max_range={adj_max}'
    if not np.isclose(np.linalg.norm(adj_min - min_range), 0, atol=1e-6) or not np.isclose(np.linalg.norm(adj_max - max_range), 0, atol=1e-6):
        Logger.warning(f"Adjusting (min_range, max_range) from ({min_range},{max_range}) to ({adj_min},{adj_max})")  # pragma: no cover


# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Callable

from model_compression_toolkit.gptq import RoundingType, GradientPTQConfigV2, GradientPTQConfig
from model_compression_toolkit.gptq.keras.quantizer.soft_rounding.soft_quantizer_reg import \
    SoftQuantizerRegularization


def get_regularization(gptq_config: GradientPTQConfig, representative_data_gen: Callable) -> Callable:
    """
    Returns a function that computes the regularization term for GPTQ training based on the given
    rounding type in the GPTQ configuration.

    Args:
        gptq_config: A GPTQ configuration.
        representative_data_gen: Dataset used for the GPTQ training.

    Returns: A function for computing the regularization. If there is no regularization function defined for the given
        rounding type, then it returns a function that just returns 0.

    """
    if gptq_config.rounding_type == RoundingType.SoftQuantizer:
        # dry run on the representative dataset to count number of batches
        num_batches = 0
        for _ in representative_data_gen():
            num_batches += 1

        n_epochs = GradientPTQConfigV2.from_v1(n_ptq_iter=num_batches, config_v1=gptq_config).n_epochs if \
            not type(gptq_config) == GradientPTQConfigV2 else gptq_config.n_epochs
        return SoftQuantizerRegularization(total_gradient_steps=num_batches * n_epochs)
    else:
        return lambda m, e_reg: 0



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import operator
from typing import Any, Callable, Dict

from model_compression_toolkit.core.common.logger import Logger


class Filter:
    """
    Filter a layer configuration by its attributes.
    """

    def match(self, layer_config: Dict[str, Any]):
        """
        Check whether the passed configuration matches the filter.
        Args:
            layer_config: Layer's configuration to check.

        Returns:
            Whether the passed configuration matches the filter or not.
        """
        raise NotImplemented('Filter did not implement match')  # pragma: no cover


class AttributeFilter(Filter):
    """
    Wrap a key, value and an operation to filter a layer's configuration according to.
    If the layer's configuration has the key, and its' value matches when applying the operator,
    the configuration matches the AttributeFilter.
    """

    def __init__(self,
                 attr: str,
                 value: Any,
                 op: Callable):
        """

        Args:
            attr (str): Attribute to filter a layer's configuration according to.
            value (Any): Value to filter to filter a layer's configuration according to.
            op (Callable): Operator to check if when applied on a layer's configuration value it holds with regard to the filter's value field.
        """
        self.attr = attr
        self.value = value
        self.op = op

    def __eq__(self, other: Any) -> bool:
        """
        Check whether an object is equal to the AttributeFilter or not.

        Args:
            other: Object to check if it is equal to the AttributeFilter or not.

        Returns:
            Whether the object is equal to the AttributeFilter or not.
        """

        if not isinstance(other, AttributeFilter):
            return False
        return self.attr == other.attr and \
               self.value == other.value and \
               self.op == other.op

    def __or__(self, other: Any):
        """
        Create a filter that combines multiple AttributeFilters with a logic OR between them.

        Args:
            other: Filter to add to self with logic OR.

        Returns:
            OrAttributeFilter that filters with OR between the current AttributeFilter and the passed AttributeFilter.
        """

        if not isinstance(other, AttributeFilter):
            Logger.error("Not an attribute filter. Can not run an OR operation.")  # pragma: no cover
        return OrAttributeFilter(self, other)

    def __and__(self, other: Any):
        """
        Create a filter that combines multiple AttributeFilters with a logic AND between them.

        Args:
            other: Filter to add to self with logic AND.

        Returns:
            AndAttributeFilter that filters with AND between the current AttributeFilter and the passed AttributeFilter.
        """
        if not isinstance(other, AttributeFilter):
            Logger.error("Not an attribute filter. Can not run an AND operation.")  # pragma: no cover
        return AndAttributeFilter(self, other)

    def match(self,
              layer_config: Dict[str, Any]) -> bool:
        """
        Check whether the passed configuration matches the filter.

        Args:
            layer_config: Layer's configuration to check.

        Returns:
            Whether the passed configuration matches the filter or not.
        """
        if self.attr in layer_config:
            return self.op(layer_config.get(self.attr), self.value)
        return False

    def op_as_str(self):
        """

        Returns: A string representation for the filter.

        """
        raise NotImplemented("Filter must implement op_as_str ")  # pragma: no cover

    def __repr__(self):
        return f'{self.attr} {self.op_as_str()} {self.value}'


class OrAttributeFilter(Filter):
    """
    AttributeFilter to filter by multiple filters with logic OR between them.
    """

    def __init__(self, *filters: AttributeFilter):
        """
        Args:
            *filters: List of filters to apply a logic OR between them when filtering.
        """
        self.filters = filters

    def match(self,
              layer_config: Dict[str, Any]) -> bool:
        """
        Check whether a layer's configuration matches the filter or not.

        Args:
            layer_config: Layer's configuration to check.

        Returns:
            Whether a layer's configuration matches the filter or not.
        """

        for f in self.filters:
            if f.match(layer_config):
                return True
        return False

    def __repr__(self):
        """

        Returns: A string representation for the filter.

        """
        return ' | '.join([str(f) for f in self.filters])


class AndAttributeFilter(Filter):
    """
    AttributeFilter to filter by multiple filters with logic AND between them.
    """

    def __init__(self, *filters):
        self.filters = filters

    def match(self,
              layer_config: Dict[str, Any]) -> bool:
        """
        Check whether the passed configuration matches the filter.
        Args:
            layer_config: Layer's configuration to check.

        Returns:
            Whether the passed configuration matches the filter or not.
        """
        for f in self.filters:
            if not f.match(layer_config):
                return False
        return True

    def __repr__(self):
        """

        Returns: A string representation for the filter.

        """
        return ' & '.join([str(f) for f in self.filters])


class Greater(AttributeFilter):
    """
    Filter configurations such that it matches configurations
    that have an attribute with a value that is greater than the value that Greater holds.
    """

    def __init__(self,
                 attr: str,
                 value: Any):
        super().__init__(attr=attr, value=value, op=operator.gt)

    def op_as_str(self): return ">"


class GreaterEq(AttributeFilter):
    """
    Filter configurations such that it matches configurations
    that have an attribute with a value that is greater or equal than the value that GreaterEq holds.
    """

    def __init__(self, attr: str, value: Any):
        super().__init__(attr=attr, value=value, op=operator.ge)

    def op_as_str(self): return ">="


class Smaller(AttributeFilter):
    """
    Filter configurations such that it matches configurations that have an attribute with a value that is smaller than the value that Smaller holds.
    """

    def __init__(self, attr: str, value: Any):
        super().__init__(attr=attr, value=value, op=operator.lt)

    def op_as_str(self): return "<"


class SmallerEq(AttributeFilter):
    """
    Filter configurations such that it matches configurations that have an attribute with a value that is smaller or equal than the value that SmallerEq holds.
    """

    def __init__(self, attr: str, value: Any):
        super().__init__(attr=attr, value=value, op=operator.le)

    def op_as_str(self): return "<="


class NotEq(AttributeFilter):
    """
    Filter configurations such that it matches configurations that have an attribute with a value that is not equal to the value that NotEq holds.
    """

    def __init__(self, attr: str, value: Any):
        super().__init__(attr=attr, value=value, op=operator.ne)

    def op_as_str(self): return "!="


class Eq(AttributeFilter):
    """
    Filter configurations such that it matches configurations that have an attribute with a value that equals to the value that Eq holds.
    """

    def __init__(self, attr: str, value: Any):
        super().__init__(attr=attr, value=value, op=operator.eq)

    def op_as_str(self): return "="


class Contains(AttributeFilter):
    """
    Filter configurations such that it matches configurations that have an attribute with a value that contains the value that Contains holds.
    """

    def __init__(self, attr: str, value: Any):
        super().__init__(attr=attr, value=value, op=operator.contains)

    def op_as_str(self): return " in "



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from enum import Enum
from typing import Dict, Any

import numpy as np


class KPITarget(Enum):
    """
    Targets for which we define KPIs metrics for mixed-precision search.
    For each target that we care to consider in a mixed-precision search, there should be defined a set of
    kpi computation function, kpi aggregation function, and kpi target (within a KPI object).

    Whenever adding a kpi metric to KPI class we should add a matching target to this enum.

    WEIGHTS - Weights memory KPI metric.

    ACTIVATION - Activation memory KPI metric.

    TOTAL - Total memory KPI metric.

    BOPS - Total Bit-Operations KPI Metric.

    """

    WEIGHTS = 'weights'
    ACTIVATION = 'activation'
    TOTAL = 'total'
    BOPS = 'bops'


class KPI:
    """
    Class to represent measurements of performance.
    """

    def __init__(self,
                 weights_memory: float = np.inf,
                 activation_memory: float = np.inf,
                 total_memory: float = np.inf,
                 bops: float = np.inf):
        """

        Args:
            weights_memory: Memory of a model's weights in bytes. Note that this includes only coefficients that should be quantized (for example, the kernel of Conv2D in Keras will be affected by this value, while the bias will not).
            activation_memory: Memory of a model's activation in bytes, according to the given activation kpi metric.
            total_memory: The sum of model's activation and weights memory in bytes, according to the given total kpi metric.
            bops: The total bit-operations in the model.
        """
        self.weights_memory = weights_memory
        self.activation_memory = activation_memory
        self.total_memory = total_memory
        self.bops = bops

    def __repr__(self):
        return f"Weights_memory: {self.weights_memory}, " \
               f"Activation_memory: {self.activation_memory}, " \
               f"Total_memory: {self.total_memory}, " \
               f"BOPS: {self.bops}"

    def get_kpi_dict(self) -> Dict[KPITarget, float]:
        """
        Returns: a dictionary with the KPI object's values for each KPI target.
        """
        return {KPITarget.WEIGHTS: self.weights_memory,
                KPITarget.ACTIVATION: self.activation_memory,
                KPITarget.TOTAL: self.total_memory,
                KPITarget.BOPS: self.bops}

    def set_kpi_by_target(self, kpis_mapping: Dict[KPITarget, float]):
        """
        Setting a KPI object values for each KPI target in the given dictionary.

        Args:
            kpis_mapping: A mapping from a KPITarget to a matching KPI value.

        """
        self.weights_memory = kpis_mapping.get(KPITarget.WEIGHTS, np.inf)
        self.activation_memory = kpis_mapping.get(KPITarget.ACTIVATION, np.inf)
        self.total_memory = kpis_mapping.get(KPITarget.TOTAL, np.inf)
        self.bops = kpis_mapping.get(KPITarget.BOPS, np.inf)

    def holds_constraints(self, kpi: Any) -> bool:
        """
        Checks whether the given KPI holds a set of KPI constraints defined by the currect KPI object.

        Args:
            kpi: A KPI object to check if it holds the constraints.

        Returns: True if all the given KPI values are not greater than the referenced KPI values.

        """
        if not isinstance(kpi, KPI):
            return False

        return kpi.weights_memory <= self.weights_memory and \
               kpi.activation_memory <= self.activation_memory and \
               kpi.total_memory <= self.total_memory and \
               kpi.bops <= self.bops



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Callable, List, Tuple, Union

import numpy as np
from torch.nn import Module
from tqdm import tqdm
import copy
import torch
from model_compression_toolkit.core.common.logger import Logger
from model_compression_toolkit.core.pytorch.back2framework.pytorch_model_builder import PyTorchModelBuilder
from model_compression_toolkit.gptq.common.gptq_graph import get_kernel_attribute_name_for_gptq
from model_compression_toolkit.gptq.common.gptq_training import GPTQTrainer
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfigV2
from model_compression_toolkit.core.common import Graph, BaseNode
from model_compression_toolkit.core.common.framework_info import FrameworkInfo
from model_compression_toolkit.core.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.core.pytorch.constants import BIAS
from model_compression_toolkit.core.pytorch.utils import to_torch_tensor, set_model, torch_tensor_to_numpy
from model_compression_toolkit.gptq.pytorch.graph_info import get_gptq_trainable_parameters, \
    get_weights_for_loss
from model_compression_toolkit.gptq.pytorch.quantizer.quantization_builder import quantization_builder
from model_compression_toolkit import quantizers_infrastructure as qi
from model_compression_toolkit.gptq.pytorch.quantizer.regularization_factory import get_regularization
from model_compression_toolkit.quantizers_infrastructure import PytorchQuantizationWrapper


class PytorchGPTQTrainer(GPTQTrainer):
    """
    Pytorch GPTQ training class for fine-tuning a quantized model
    """

    def __init__(self,
                 graph_float: Graph,
                 graph_quant: Graph,
                 gptq_config: GradientPTQConfigV2,
                 fw_impl: FrameworkImplementation,
                 fw_info: FrameworkInfo,
                 representative_data_gen: Callable):
        """
        Build two models from a graph: A teacher network (float model) and a student network (quantized model).
        Use the dataset generator to pass images through the teacher and student networks to get intermediate
        layers outputs. Use the outputs to compute the observed loss and to back-propagate the error
        in the student network, to minimize it in the next similar steps.
        All parameters (such as number of iterations, optimizer, etc.) are in GradientPTQConfig.
        Args:
            graph_float: Graph to build a float networks from.
            graph_quant: Graph to build a quantized networks from.
            gptq_config: GradientPTQConfigV2 with parameters about the tuning process.
            fw_impl: FrameworkImplementation object with a specific framework methods implementation.
            fw_info: Framework information
            representative_data_gen: Dataset to use for inputs of the models.
        """
        super().__init__(graph_float, graph_quant, gptq_config, fw_impl, fw_info)
        self.loss_list = []
        self.input_scale = 1
        if self.float_user_info.input_scale != self.gptq_user_info.input_scale:
            Logger.error("Input scale mismatch between float and GPTQ networks")  # pragma: no cover
        else:
            self.input_scale = self.gptq_user_info.input_scale

        trainable_weights, trainable_bias, trainable_threshold = get_gptq_trainable_parameters(
            self.fxp_model,
            add_bias=self.gptq_config.train_bias)

        self.flp_weights_list, self.fxp_weights_list = get_weights_for_loss(self.fxp_model)
        if not (len(self.compare_points) == len(trainable_weights) == len(self.flp_weights_list) == len(
                self.fxp_weights_list)):
            Logger.error(
                "GPTQ: Mismatch between number of compare points, number of layers with trainable weights " +
                "and number of float and quantized weights for loss")

        self.optimizer_with_param = self.get_optimizer_with_param(trainable_weights,
                                                                  trainable_bias,
                                                                  trainable_threshold)

        self.weights_for_average_loss = to_torch_tensor(self.compute_hessian_based_weights(representative_data_gen))

        self.reg_func = get_regularization(self.gptq_config, representative_data_gen)

    def _is_gptq_applicable(self,
                            node: BaseNode) -> bool:
        """
        A function for deciding if a layer should be fine-tuned during GPTQ.
        Args:
            node (BaseNode): Node for quantization decision
        Returns:
            A boolean whether the layer is to be wrapped with a Quantization Wrapper.
        """

        if node.is_weights_quantization_enabled() and not self.fw_info.is_kernel_op(node.type):
            Logger.error(f"GPTQ Error: Quantizing node {node.name} of type {node.type} "
                         f"without a kernel isn't supported.")
        return node.is_weights_quantization_enabled()

    def gptq_wrapper(self, n: BaseNode, layer: Module) -> Union[qi.PytorchQuantizationWrapper, Module]:
        """
        A function which takes a computational graph node and a pytorch layer and perform the quantization wrapping.

        Args:
            n: A node of mct graph.
            layer: A pytorch layer

        Returns: Wrapped layer if the layer should be wrap, otherwise returns the layer as is.
        """

        if self._is_gptq_applicable(n):
            weights_quantizers, activation_quantizers = quantization_builder(n, self.gptq_config)
            return qi.PytorchQuantizationWrapper(layer,
                                                 weights_quantizers=weights_quantizers,
                                                 activation_quantizers=activation_quantizers)
        else:
            return layer

    def build_gptq_model(self):
        """
        Build the GPTQ model with QuantizationWrappers
        Returns:
            Quantized graph for GPTQ fine-tuning, GPTQ graph user info
        """
        gptq_model, gptq_user_info = PyTorchModelBuilder(graph=self.graph_quant,
                                                         append2output=self.compare_points,
                                                         fw_info=self.fw_info,
                                                         wrapper=self.gptq_wrapper,
                                                         return_float_outputs=True).build_model()

        return gptq_model, gptq_user_info

    def train(self, representative_data_gen: Callable):
        """
          GPTQ Training using pytorch framework
          Args:
              representative_data_gen: Dataset generator to get images.
          Returns:
              Graph after GPTQ training
          """
        # Set Optimizers
        for (optimizer, params) in self.optimizer_with_param:
            optimizer.param_groups.clear()
            optimizer.add_param_group({'params': params})

        # Set models mode
        set_model(self.float_model, False)
        set_model(self.fxp_model, True)
        self._set_requires_grad()

        # ----------------------------------------------
        # Training loop
        # ----------------------------------------------
        self.micro_training_loop(representative_data_gen, self.gptq_config.n_epochs)

    def compute_gradients(self,
                          y_float: List[torch.Tensor],
                          input_tensors: List[torch.Tensor]) -> Tuple[torch.Tensor, List[np.ndarray]]:
        """
        Get outputs from both teacher and student networks. Compute the observed error,
        and use it to compute the gradients and applying them to the student weights.
        Args:
            y_float: A list of reference tensor from the floating point network.
            input_tensors: A list of Input tensors to pass through the networks.
        Returns:
            Loss and gradients.
        """

        # Forward-pass
        y_fxp = self.fxp_model(input_tensors)

        # Loss
        loss_value = self.gptq_config.loss(y_fxp,
                                           y_float,
                                           self.fxp_weights_list,
                                           self.flp_weights_list,
                                           self.compare_points_mean,
                                           self.compare_points_std,
                                           self.weights_for_average_loss)

        reg_value = self.reg_func(self.fxp_model, self.gptq_config.regularization_factor)

        loss_value += reg_value

        # Back-pass
        loss_value.backward()

        # Get gradients
        grads = []
        for param in self.fxp_model.parameters():
            if param.requires_grad and param.grad is not None:
                grads.append(torch_tensor_to_numpy(param.grad))

        return loss_value, grads

    def micro_training_loop(self,
                            data_function: Callable,
                            n_epochs: int):
        """
        This function run a micro training loop on given set of parameters.
        Args:
            data_function: A callable function that give a batch of samples.
            n_epochs: Number of update iterations of representative dataset.
        """
        for _ in tqdm(range(n_epochs)):
            for data in tqdm(data_function()):
                input_data = [d * self.input_scale for d in data]
                input_tensor = to_torch_tensor(input_data)
                y_float = self.float_model(input_tensor)  # running float model
                loss_value, grads = self.compute_gradients(y_float, input_tensor)
                # Run one step of gradient descent by updating the value of the variables to minimize the loss.
                for (optimizer, _) in self.optimizer_with_param:
                    optimizer.step()
                    optimizer.zero_grad()
                if self.gptq_config.log_function is not None:
                    self.gptq_config.log_function(loss_value.item(),
                                                  torch_tensor_to_numpy(grads),
                                                  torch_tensor_to_numpy(self.optimizer_with_param[0][-1]))
                self.loss_list.append(loss_value.item())
                Logger.debug(f'last loss value: {self.loss_list[-1]}')

    def update_graph(self) -> Graph:
        """
        Update a graph using GPTQ after minimizing the loss between the float model's output
        and the quantized model's outputs.
        Returns:
            Updated graph after GPTQ.
        """
        graph_quant = copy.copy(self.graph_quant)

        # Update graph after training
        for name, layer in self.fxp_model.named_modules():
            if isinstance(layer, PytorchQuantizationWrapper):
                node = self.graph_quant.find_node_by_name(name)
                if len(node) != 1:
                    Logger.error(f"Can't update GPTQ graph due to missing layer named: {name}")
                node = node[0]
                kernel_attribute = get_kernel_attribute_name_for_gptq(layer_type=node.type,
                                                                      fw_info=self.fw_info)
                weights, weight_quant_config, activation_quant_config = \
                    layer.weights_quantizers[kernel_attribute].update_layer_quantization_params(layer)
                for weight_attr, weight in weights.items():
                    node.set_weights_by_keys(weight_attr, self.fw_impl.to_numpy(weight))
                for config_attr, config_value in weight_quant_config.items():
                    node.final_weights_quantization_cfg.set_quant_config_attr(config_attr, config_value)
                for config_attr, config_value in activation_quant_config.items():
                    node.final_activation_quantization_cfg.set_quant_config_attr(config_attr, config_value)
                if self.gptq_config.train_bias and hasattr(layer.layer, BIAS):
                    node.set_weights_by_keys(BIAS, self.fw_impl.to_numpy(getattr(layer.layer, BIAS)))

        return graph_quant

    def _set_requires_grad(self):
        """
        Set require_grad flag for trainable parameters for GPTQ training
        """
        # Float model: freeze all the parameters in the network
        for param in self.float_model.parameters():
            param.requires_grad = False

        # Fxp model: unfreeze bias trainable parameters
        for layer in self.fxp_model.modules():
            if isinstance(layer, PytorchQuantizationWrapper):
                if hasattr(layer.layer, BIAS):
                    bias = getattr(layer.layer, BIAS)
                    bias.requires_grad = self.gptq_config.train_bias



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from typing import Callable

from model_compression_toolkit.core.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.core.common import FrameworkInfo, Logger
from model_compression_toolkit.core.common.constants import NUM_SAMPLES_DISTANCE_TENSORBOARD
from model_compression_toolkit.core.common.graph.base_graph import Graph

from model_compression_toolkit.core.common.similarity_analyzer import compute_cs
from model_compression_toolkit.core.common.visualization.nn_visualizer import NNVisualizer

from model_compression_toolkit.core.common.visualization.tensorboard_writer import TensorboardWriter


def analyzer_model_quantization(representative_data_gen: Callable,
                                tb_w: TensorboardWriter,
                                tg: Graph,
                                fw_impl: FrameworkImplementation,
                                fw_info: FrameworkInfo):
    """
    Plot the cosine similarity of different points on the graph between the float and quantized
    graphs. Add them to the passed TensorboardWriter object and close all tensorboard writer open
    files.

    Args:
        representative_data_gen: Dataset used for calibration.
        tb_w: TensorBoardWriter object to log events.
        tg: Graph of quantized model.
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.
        fw_info: Information needed for quantization about the specific framework.

    """
    if tb_w is not None:
        visual = NNVisualizer(tg,
                              fw_impl=fw_impl,
                              fw_info=fw_info)

        for i, _data in enumerate(representative_data_gen()):
            if i >= NUM_SAMPLES_DISTANCE_TENSORBOARD:
                break
            figure = visual.plot_distance_graph(_data,
                                                distance_fn=compute_cs,
                                                convert_to_range=lambda a: 1 - 2 * a)
            tb_w.add_figure(figure, f'similarity_distance_sample_{i}')
        else:
            Logger.warning(f'Not enough batches in representative dataset to generate {NUM_SAMPLES_DISTANCE_TENSORBOARD} figures')
        tb_w.close()
